# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Version of anaconda-project library."""

VERSION_INFO = (0, 8, 3)
__version__ = '.'.join(map(str, VERSION_INFO))
version = __version__
